import { Observable } from 'rxjs';
import { Arc, IApolloQueryOptions } from './arc';
import { IGenesisProtocolParams } from './genesisProtocol';
import { ITransaction, Operation, transactionErrorHandler, transactionResultHandler } from './operation';
import { IProposalCreateOptions, IProposalQueryOptions, Proposal } from './proposal';
import { SchemeBase } from './schemes/base';
import { ReputationFromTokenScheme } from './schemes/reputationFromToken';
import { Address, ICommonQueryOptions } from './types';
export interface ISchemeState {
    id: string;
    address: Address;
    dao: Address;
    name: string;
    version: string;
    canDelegateCall: boolean;
    canRegisterSchemes: boolean;
    canUpgradeController: boolean;
    canManageGlobalConstraints: boolean;
    contributionRewardParams?: IContributionRewardParams;
    contributionRewardExtParams?: IContributionRewardExtParams;
    genericSchemeParams?: IGenericSchemeParams;
    schemeRegistrarParams?: {
        votingMachine: Address;
        voteRemoveParams: IGenesisProtocolParams;
        voteRegisterParams: IGenesisProtocolParams;
    } | null;
    numberOfQueuedProposals: number;
    numberOfPreBoostedProposals: number;
    numberOfBoostedProposals: number;
    schemeParams?: IGenericSchemeParams | IContributionRewardParams | IContributionRewardExtParams | ISchemeRegisterParams;
}
export interface IGenericSchemeParams {
    votingMachine: Address;
    contractToCall: Address;
    voteParams: IGenesisProtocolParams;
}
export interface IContributionRewardParams {
    votingMachine: Address;
    voteParams: IGenesisProtocolParams;
}
export interface IContributionRewardExtParams {
    votingMachine: Address;
    voteParams: IGenesisProtocolParams;
    rewarder: Address;
}
export interface ISchemeRegisterParams {
    votingMachine: Address;
    contractToCall: Address;
    voteParams: IGenesisProtocolParams;
}
export interface ISchemeQueryOptions extends ICommonQueryOptions {
    where?: {
        address?: Address;
        canDelegateCall?: boolean;
        canRegisterSchemes?: boolean;
        canUpgradeController?: boolean;
        canManageGlobalConstraints?: boolean;
        dao?: Address;
        id?: string;
        name?: string;
        [key: string]: any;
    };
}
export interface ISchemeQueryOptions extends ICommonQueryOptions {
    where?: {
        address?: Address;
        canDelegateCall?: boolean;
        canRegisterSchemes?: boolean;
        canUpgradeController?: boolean;
        canManageGlobalConstraints?: boolean;
        dao?: Address;
        id?: string;
        name?: string;
        [key: string]: any;
    };
}
/**
 * A Scheme represents a scheme instance that is registered at a DAO
 */
export declare class Scheme extends SchemeBase {
    context: Arc;
    /**
     * Scheme.search(context, options) searches for scheme entities
     * @param  context an Arc instance that provides connection information
     * @param  options the query options, cf. ISchemeQueryOptions
     * @return         an observable of Scheme objects
     */
    static search(context: Arc, options?: ISchemeQueryOptions, apolloQueryOptions?: IApolloQueryOptions): Observable<SchemeBase[]>;
    /**
     * map an apollo query result to ISchemeState
     *
     * @static
     * @param {*} item
     * @param {Arc} arc
     * @returns {(ISchemeState|null)}
     * @memberof Scheme
     */
    static itemMap(arc: Arc, item: any): ISchemeState | null;
    id: Address;
    coreState: ISchemeState | null;
    ReputationFromToken: ReputationFromTokenScheme | null;
    constructor(context: Arc, idOrOpts: Address | ISchemeState);
    setState(opts: ISchemeState): void;
    /**
     * fetch the static state from the subgraph
     * @return the statatic state
     */
    fetchState(apolloQueryOptions?: IApolloQueryOptions): Promise<ISchemeState>;
    state(apolloQueryOptions?: IApolloQueryOptions): Observable<ISchemeState>;
    /**
     * create a new proposal in this Scheme
     * @param  options [description ]
     * @return a Proposal instance
     */
    createProposal(options: IProposalCreateOptions): Operation<Proposal>;
    proposals(options?: IProposalQueryOptions, apolloQueryOptions?: IApolloQueryOptions): Observable<Proposal[]>;
    protected createProposalTransaction(options: IProposalCreateOptions): Promise<ITransaction>;
    protected createProposalTransactionMap(): transactionResultHandler<Proposal>;
    protected createProposalErrorHandler(options: IProposalCreateOptions): transactionErrorHandler;
}
