import { Observable } from 'rxjs';
import { Arc, IApolloQueryOptions } from '../arc';
import { IGenesisProtocolParams } from '../genesisProtocol';
import { ITransaction, Operation, transactionErrorHandler, transactionResultHandler } from '../operation';
import { IProposalCreateOptions, IProposalQueryOptions, Proposal } from '../proposal';
import { Address, ICommonQueryOptions, IStateful } from '../types';
import { ReputationFromTokenScheme } from './reputationFromToken';
export interface ISchemeState {
    id: string;
    address: Address;
    dao: Address;
    name: string;
    version: string;
    canDelegateCall: boolean;
    canRegisterSchemes: boolean;
    canUpgradeController: boolean;
    canManageGlobalConstraints: boolean;
    contributionRewardParams?: IContributionRewardParams;
    contributionRewardExtParams?: IContributionRewardExtParams;
    genericSchemeParams?: IGenericSchemeParams;
    schemeRegistrarParams?: {
        votingMachine: Address;
        voteRemoveParams: IGenesisProtocolParams;
        voteRegisterParams: IGenesisProtocolParams;
    } | null;
    numberOfQueuedProposals: number;
    numberOfPreBoostedProposals: number;
    numberOfBoostedProposals: number;
    schemeParams?: IGenericSchemeParams | IContributionRewardParams | IContributionRewardExtParams | ISchemeRegisterParams;
}
export interface IGenericSchemeParams {
    votingMachine: Address;
    contractToCall: Address;
    voteParams: IGenesisProtocolParams;
}
export interface IContributionRewardParams {
    votingMachine: Address;
    voteParams: IGenesisProtocolParams;
}
export interface IContributionRewardExtParams {
    votingMachine: Address;
    voteParams: IGenesisProtocolParams;
    rewarder: Address;
}
export interface ISchemeRegisterParams {
    votingMachine: Address;
    contractToCall: Address;
    voteParams: IGenesisProtocolParams;
}
export interface ISchemeQueryOptions extends ICommonQueryOptions {
    where?: {
        address?: Address;
        canDelegateCall?: boolean;
        canRegisterSchemes?: boolean;
        canUpgradeController?: boolean;
        canManageGlobalConstraints?: boolean;
        dao?: Address;
        id?: string;
        name?: string;
        [key: string]: any;
    };
}
export interface ISchemeQueryOptions extends ICommonQueryOptions {
    where?: {
        address?: Address;
        canDelegateCall?: boolean;
        canRegisterSchemes?: boolean;
        canUpgradeController?: boolean;
        canManageGlobalConstraints?: boolean;
        dao?: Address;
        id?: string;
        name?: string;
        [key: string]: any;
    };
}
/**
 * A Scheme represents a scheme instance that is registered at a DAO
 */
export declare abstract class SchemeBase implements IStateful<ISchemeState> {
    context: Arc;
    static fragments: {
        SchemeFields: import("graphql").DocumentNode;
    };
    id: Address;
    coreState: ISchemeState | null;
    ReputationFromToken: ReputationFromTokenScheme | null;
    constructor(context: Arc, idOrOpts: Address | ISchemeState);
    /**
     * fetch the static state from the subgraph
     * @return the statatic state
     */
    fetchState(apolloQueryOptions?: IApolloQueryOptions): Promise<ISchemeState>;
    setState(opts: ISchemeState): void;
    createProposal(options: IProposalCreateOptions): Operation<Proposal>;
    abstract state(apolloQueryOptions?: IApolloQueryOptions): Observable<ISchemeState>;
    proposals(options?: IProposalQueryOptions, apolloQueryOptions?: IApolloQueryOptions): Observable<Proposal[]>;
    /**
     * create a new proposal in this scheme
     * TODO: move this to the schemes - we should call proposal.scheme.createProposal
     * @param  options [description ]
     * @return a Proposal instance
     */
    protected abstract createProposalTransaction(options: IProposalCreateOptions): Promise<ITransaction>;
    protected abstract createProposalTransactionMap(): transactionResultHandler<Proposal>;
    protected abstract createProposalErrorHandler(options: IProposalCreateOptions): transactionErrorHandler;
}
